package de.fhdw.wtf.common.ast.type;

import de.fhdw.wtf.common.ast.Name;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorReturnException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.exception.ast.ASTException;
import de.fhdw.wtf.common.token.Token;

/**
 * This class represents all available baseTypes for the code generation. All available baseTypes are specified in
 * {@link de.fhdw.wtf.common.constants.referencer.BaseTypeConstants}.
 */
public final class BaseType extends AtomicType {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = 1931081272975105826L;
	
	/**
	 * Constant to define hashCode().
	 */
	private static final int HASHCODE_CONSTANT = 23123234;
	
	/**
	 * Constructor of {@link BaseType}.
	 * 
	 * @param typeName
	 *            typeName
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 */
	private BaseType(final Name typeName, final Token firstToken, final Token lastToken) {
		super(typeName, firstToken, lastToken);
	}
	
	/**
	 * Create a {@link BaseType}-Object.
	 * 
	 * @param typeName
	 *            typeName
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 * @return The BaseType-Object
	 */
	public static BaseType create(final Name typeName, final Token firstToken, final Token lastToken) {
		return new BaseType(typeName, firstToken, lastToken);
	}
	
	@Override
	public boolean equals(final Object o) {
		final boolean resultSuper = super.equals(o);
		return resultSuper && o instanceof BaseType;
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ HASHCODE_CONSTANT;
	}
	
	@Override
	public String toString() {
		return this.getTypeName().toString();
	}
	
	@Override
	public String getTypeString() {
		return this.toString();
	}
	
	@Override
	public int compareTo(final Type o) {
		return o.accept(new TypeVisitorReturn<Integer>() {
			
			@Override
			public Integer handle(final AtomicType s) {
				return s.accept(new AtomicTypeVisitorReturn<Integer>() {
					
					@Override
					public Integer handle(final BaseType baseType) {
						return BaseType.this.toString().compareTo(baseType.toString());
					}
					
					@Override
					public Integer handle(final ClassType clazz) {
						return -1;
					}
				});
			}
			
			@Override
			public Integer handle(final CompositeType c) {
				return -1;
			}
			
			@Override
			public Integer handle(final TypeProxy s) {
				return 1;
			}
		});
	}
	
	@Override
	public void accept(final AtomicTypeVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final AtomicTypeVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	@Override
	public <Y extends Exception> void accept(final AtomicTypeVisitorException<Y> v) throws Y {
		v.handle(this);
	}
	
	@Override
	public <X, Y extends Exception> X accept(final AtomicTypeVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
	
	@Override
	public boolean isSupertypeOf(final Type type) {
		return false;
	}
	
	@Override
	public BaseType getPrototype() {
		return (BaseType) super.getPrototype();
	}
	
	@Override
	public void setPrototype(final Type prototype) {
		if (!(prototype instanceof BaseType)) {
			throw new ASTException("The prototype \"" + prototype.getClass() + "\" does not match this type \""
					+ this.getClass() + "\"!");
		}
		super.setPrototype(prototype);
	}
	
}
