package de.fhdw.wtf.common.ast.type;

import java.util.Collection;
import java.util.List;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.Constructor;
import de.fhdw.wtf.common.ast.Name;
import de.fhdw.wtf.common.ast.Operation;
import de.fhdw.wtf.common.ast.visitor.ClassTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.ClassTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.ClassTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.ClassTypeVisitorReturnException;
import de.fhdw.wtf.common.token.Token;

/**
 * This class represents a class for code generation. Such a {@link RegularClassType} is always held by a <Group>, can
 * contain multiple {@link ClassModifier}s , multiple {@link Attribute}s, multiple {@link Operation}s and extend other
 * {@link RegularClassType}s.
 */
public class RegularClassType extends ClassType {
	
	/**
	 * Constructor of {@link RegularClassType}.
	 * 
	 * @param name
	 *            The name of the {@link RegularClassType}.
	 * @param modifiers
	 *            The modifiers of the {@link RegularClassType}.
	 * @param attributes
	 *            The attributes of the {@link RegularClassType}.
	 * @param superTypes
	 *            The supertypes of the {@link RegularClassType}.
	 * @param operations
	 *            The operations of the {@link RegularClassType}.
	 * @param constructors
	 *            The constructors called by {@link RegularClassType}.
	 * @param firstToken
	 *            firstToken
	 * @param subTypes
	 *            The subtypes of the {@link RegularClassType}.
	 * @param lastToken
	 *            lastToken
	 */
	private RegularClassType(final Name name,
			final Collection<ClassModifier> modifiers,
			final Collection<Attribute> attributes,
			final List<Type> superTypes,
			final Collection<Operation> operations,
			final Collection<Constructor> constructors,
			final Token firstToken,
			final Collection<ClassType> subTypes,
			final Token lastToken) {
		super(name, modifiers, attributes, superTypes, operations, constructors, firstToken, subTypes, lastToken);
		
	}
	
	/**
	 * Creates a {@link RegularClassType}-Object.
	 * 
	 * @param name
	 *            The name of the {@link RegularClassType}.
	 * @param modifiers
	 *            The modifiers of the {@link RegularClassType}.
	 * @param attributes
	 *            The attributes of the {@link RegularClassType}.
	 * @param superTypes
	 *            The supertypes of the {@link RegularClassType}.
	 * @param operations
	 *            The operations of the {@link RegularClassType}.
	 * @param superConstructors
	 *            The constructors called by {@link RegularClassType}.
	 * @param firstToken
	 *            firstToken
	 * @param subTypes
	 *            The subtypes of the {@link RegularClassType}.
	 * @param lastToken
	 *            lastToken
	 * @return The {@link RegularClassType}-Object.
	 */
	public static RegularClassType create(final Name name,
			final Collection<ClassModifier> modifiers,
			final Collection<Attribute> attributes,
			final List<Type> superTypes,
			final Collection<Operation> operations,
			final Collection<Constructor> superConstructors,
			final Token firstToken,
			final Collection<ClassType> subTypes,
			final Token lastToken) {
		return new RegularClassType(name, modifiers, attributes, superTypes, operations, superConstructors, firstToken,
				subTypes, lastToken);
	}
	
	/**
	 * Creates a {@link RegularClassType}-Object.
	 * 
	 * @param name
	 *            The name of the {@link RegularClassType}.
	 * @param modifiers
	 *            The modifiers of the {@link RegularClassType}.
	 * @param attributes
	 *            The attributes of the {@link RegularClassType}.
	 * @param superTypes
	 *            The supertypes of the {@link RegularClassType}.
	 * @param operations
	 *            The operations of the {@link RegularClassType}.
	 * @param superConstructors
	 *            The constructors called by {@link RegularClassType}.
	 * @param firstToken
	 *            firstToken
	 * @param subTypes
	 *            The subtypes of the {@link RegularClassType}.
	 * @return The {@link RegularClassType}-Object.
	 */
	public static RegularClassType create(final Name name,
			final Collection<ClassModifier> modifiers,
			final Collection<Attribute> attributes,
			final List<Type> superTypes,
			final Collection<Operation> operations,
			final Collection<Constructor> superConstructors,
			final Token firstToken,
			final Collection<ClassType> subTypes) {
		return new RegularClassType(name, modifiers, attributes, superTypes, operations, superConstructors, firstToken,
				subTypes, null);
	}
	
	@Override
	public void accept(final ClassTypeVisitor classTypeVisitor) {
		classTypeVisitor.handle(this);
	}
	
	@Override
	public <X> X accept(final ClassTypeVisitorReturn<X> classTypeVisitorReturn) {
		return classTypeVisitorReturn.handle(this);
	}
	
	@Override
	public <Y extends Exception> void accept(final ClassTypeVisitorException<Y> classTypeVisitorException) throws Y {
		classTypeVisitorException.handle(this);
	}
	
	@Override
	public <X, Y extends Exception> X accept(final ClassTypeVisitorReturnException<X, Y> classTypeVisitorReturnException)
			throws Y {
		return classTypeVisitorReturnException.handle(this);
	}
}
