package de.fhdw.wtf.common.task;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import de.fhdw.wtf.common.task.result.TaskResult;

/**
 * 
 *
 */
public final class TaskExecutorFixed implements TaskExecutor {
	
	private static final int MAX_NUM_THREADS = 10;
	
	private final ExecutorService executor;
	
	private final ConcurrentLinkedQueue<Future<TaskResult>> results;
	
	private final Collection<DependencyTask> knownTasks;
	
	/**
	 * Constructor of {@link TaskExecutorFixed}.
	 */
	private TaskExecutorFixed() {
		this.executor = Executors.newFixedThreadPool(MAX_NUM_THREADS);
		this.results = new ConcurrentLinkedQueue<>();
		this.knownTasks = new ArrayList<>();
	}
	
	/**
	 * Creates a {@link TaskExecutorFixed}-Object.
	 * 
	 * @return The {@link TaskExecutorFixed}-Object.
	 */
	public static TaskExecutorFixed create() {
		return new TaskExecutorFixed();
	}
	
	/**
	 * Starts all tasks contained by the given collection <code>tasks</code>. Successors of this tasks are specified by
	 * its dependencies.
	 * 
	 * @param tasks
	 *            : Collection of Tasks.
	 */
	public void start(final Collection<? extends DependencyTask> tasks) {
		for (final DependencyTask task : tasks) {
			this.submit(task);
		}
	}
	
	@Override
	public synchronized void submit(final DependencyTask a) {
		final Future<TaskResult> result = this.executor.submit(a);
		this.results.add(result);
	}
	
	private synchronized boolean isQueueEmpty() {
		return this.results.isEmpty();
	}
	
	@Override
	public Collection<TaskResult> getResultsAndShutdown() throws InterruptedException, ExecutionException {
		final Collection<TaskResult> result = new ArrayList<>();
		while (!this.isQueueEmpty()) {
			final Future<TaskResult> current = this.results.poll();
			result.add(current.get());
		}
		this.executor.shutdown();
		return result;
	}
	
	@Override
	public void introduce(final DependencyTask newTask) {
		this.knownTasks.add(newTask);
	}
	
	@Override
	public void startAllKnownTasks() throws InterruptedException {
		final Collection<DependencyTask> startThese = new ArrayList<>();
		final Iterator<DependencyTask> i = this.knownTasks.iterator();
		while (i.hasNext()) {
			final DependencyTask current = i.next();
			if (current.getDependencies().isEmpty()) {
				startThese.add(current);
			}
		}
		this.results.addAll(this.executor.invokeAll(startThese));
	}
	
}
